# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

"""batou templating support

Currently we support two templating engines with different formats.

Mako::
    % for server in servers:
    server ${server.name}
    % endfor

Jinja2::
    %% for server in servers
    server {{ server.name }}
    %% endfor
"""

from __future__ import print_function, unicode_literals
import jinja2
import mako.runtime
import mako.template
import StringIO


class TemplateEngine(object):
    """Abstract templating wrapper class.

    Use a subclass that connects to a specific template engine.
    """

    @classmethod
    def get(cls, enginename):
        """Return TemplateEngine instance for `enginename`."""
        if enginename.lower() == 'mako':
            return MakoEngine()
        if enginename.lower() == 'jinja2':
            return Jinja2Engine()
        raise NotImplementedError('template engine not known', enginename)

    def template(self, sourcefile, targetfile, args):
        """Render template from `sourcefile` to `targetfile`.

        Return True if the targetfile has actually been changed.
        """
        interpolated = self._render_template_file(sourcefile, args)
        try:
            with open(targetfile, 'r') as old:
                if old.read() == interpolated.getvalue():
                    return False
        except EnvironmentError:
            pass
        with open(targetfile, 'w') as output:
            output.write(interpolated.getvalue())
        return True

    def _render_template_file(sourcefile, args):
        """Expand template found in `sourcefile` and return it as StringIO."""
        raise NotImplementedError

    def expand(self, templatestr, args):
        """Expand template in `templatestr` and return it as string."""
        raise NotImplementedError


class MakoEngine(TemplateEngine):

    def _render_template_file(self, sourcefile, args):
        args.setdefault('d', '$')
        output = StringIO.StringIO()
        ctx = mako.runtime.Context(output, **args)
        tmpl = mako.template.Template(filename=sourcefile,
                                      strict_undefined=True)
        tmpl.render_context(ctx)
        return output

    def expand(self, templatestr, args):
        args.setdefault('d', '$')
        tmpl = mako.template.Template(templatestr)
        return tmpl.render(**args)


class Jinja2Engine(TemplateEngine):

    def __init__(self, *args, **kwargs):
        super(Jinja2Engine, self).__init__(*args, **kwargs)
        self.env = jinja2.Environment(
            line_statement_prefix='@@',
            undefined=jinja2.StrictUndefined,
            loader=jinja2.FileSystemLoader(['.', '/']))

    def _render_template_file(self, sourcefile, args):
        tmpl = self.env.get_template(sourcefile)
        output = StringIO.StringIO()
        print(tmpl.render(args), file=output)
        return output

    def expand(self, templatestr, args):
        tmpl = self.env.from_string(templatestr)
        return tmpl.render(**args)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

import os
import pwd


class Environment(object):
    """An environment is an allocation of components to hosts."""

    service_user = pwd.getpwuid(os.getuid()).pw_name
    host_domain = None
    branch = u'default'
    passphrase_file = None

    def __init__(self, name, service):
        self.name = name
        self.service = service
        self.hosts = {}

    def configure(self, config):
        """Pull options that come from cfg file out of `config` dict."""
        self.service_user = config.get('service_user', self.service_user)
        self.host_domain = config.get('host_domain', self.host_domain)
        self.branch = config.get('branch', self.branch)

    def normalize_host_name(self, hostname):
        """Ensure the given host name is an FQDN for this environment."""
        if not self.host_domain:
            return hostname
        domain = self.host_domain
        return '%s.%s' % (hostname.rstrip(domain), domain)

    def get_host(self, hostname):
        return self.hosts[self.normalize_host_name(hostname)]



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from batou.passphrase import PassphraseFile
import atexit
import fabric.api
import fabric.contrib.files
import fabric.main
import os
import re
import subprocess


def main():
    # Bootstrap this batou project on the remote hosts and invoke local channel
    # deployments there.
    # This fabric call will turn around, use our stub fabfile which will then
    # in turn use the FabricTask wrapper to bootstrap remote hosts and
    # invoke us over there.
    fabric.main.main()


class FabricTask(fabric.tasks.Task):
    """Adapter to bind fabric's task API and our environment/hosts together."""

    def __init__(self, config, name):
        self.config = config
        self.service = config.service
        self.environment = config.service.environments[name]
        self.name = name

    @classmethod
    def from_config(cls, config):
        tasks = {}
        for name in config.service.environments:
            tasks[name] = task = cls(config, name)
            task.__doc__ = 'Deploy to %s environment' % name
        return tasks

    def get_hosts(self, arg_hosts, arg_roles, arg_exclude_hosts, env=None):
        """Compute effective subset if hosts have been named on commandline."""
        allowed_hosts = set(self.environment.hosts)
        if arg_hosts:
            arg_hosts = set((self.environment.normalize_host_name(h)
                             for h in arg_hosts))
            return set(arg_hosts).intersection(allowed_hosts)
        return allowed_hosts

    def setup_passphrase(self):
        """Set up passphrase locally."""
        self._pf = PassphraseFile(u'environment "%s"' % self.environment.name)
        self.passphrase_file = self._pf.__enter__()
        atexit.register(self._pf.__exit__)

    def transfer_passphrase(self):
        """Set up passphrase on remote host in .batou-passphrase"""
        fabric.api.put(self.passphrase_file, u'/tmp/batou-passphrase')
        self.cmd(u'install -m0600 /tmp/batou-passphrase %s/.batou-passphrase' %
                 self.remote_base)
        fabric.api.run(u'rm -f /tmp/batou-passphrase')

    def bouncedir_name(self):
        """Pick suitable name for hg repository bounce dir."""
        raw_name = u'bounce-%s-%s-%s' % (
            os.path.dirname(__file__).rsplit(u'/')[-1], self.environment.name,
            self.environment.branch)
        return re.sub(r'[^a-zA-Z0-9_.-]', u'_', raw_name)

    def bootstrap(self):
        """Ensure that the batou code base is current."""
        bouncedir = fabric.api.run(u'echo -n ~/%s' % self.bouncedir_name())
        if not self.exists(bouncedir):
            fabric.api.run(u'hg init %s' % bouncedir)
        try:
            subprocess.check_call(['hg push -f ssh://%s/%s' %
                                   (fabric.api.env.host, bouncedir)],
                                  shell=True)
        except subprocess.CalledProcessError, e:
            if e.returncode != 1:
                # 1 means: nothing to push
                raise
        if not self.exists(self.remote_base):
            self.cmd(u'hg clone %s %s' % (bouncedir, self.remote_base))
        else:
            with self.cd(self.remote_base):
                self.cmd(u'hg pull %s' % bouncedir)
        self.transfer_passphrase()
        with self.cd(self.remote_base):
            self.cmd(u'hg update -C %s' % self.environment.branch)
            if not self.exists('bin/python2.7'):
                self.cmd('virtualenv --no-site-packages --python python2.7 .')
            # XXX We're having upgrade issues with setuptools. We always
            # bootstrap, for now.
            self.cmd('bin/python2.7 bootstrap.py')
            self.cmd('bin/buildout -t 15')

    def run(self):
        """This is fabric's `run`. We would call this 'deploy'."""
        self.setup_passphrase()
        # Lazy loading of the config to limit this to the environment we
        # actually need.
        self.environment = self.config.service.environments[self.name]
        self.environment.passphrase_file = self.passphrase_file
        self.name = self.environment.name
        self.remote_base = '~%s/deployment' % self.environment.service_user
        self.config.configure_components(self.environment)

        host = self.environment.hosts[fabric.api.env.host]
        self.bootstrap()
        with fabric.api.cd(self.remote_base):
            # Run local batou deployment on remote host
            self.cmd('bin/batou-local %s %s' %
                     (self.environment.name, host.fqdn))

    # Fabric convenience

    def cmd(self, cmd):
        """Execute `cmd` in the remote service user's context."""
        return fabric.api.sudo(cmd, user=self.environment.service_user)

    def cd(self, path):
        return fabric.api.cd(path)

    def exists(self, path):
        return fabric.contrib.files.exists(path)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from batou.environment import Environment
import unittest


class EnvironmentTest(unittest.TestCase):

    def test_configure_should_use_defaults(self):
        e = Environment(u'name', u'service')
        e.configure({})
        self.assertEqual(Environment.service_user, e.service_user)
        self.assertEqual(Environment.host_domain, e.host_domain)
        self.assertEqual(Environment.branch, e.branch)

    def test_configure_should_use_config(self):
        e = Environment(u'name', u'service')
        e.configure(dict(
            service_user=u'joe',
            host_domain=u'example.com',
            branch=u'release'
        ))
        self.assertEqual(u'joe', e.service_user)
        self.assertEqual(u'example.com', e.host_domain)
        self.assertEqual(u'release', e.branch)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

"""Query passphrase from user and put it into a temporary file"""

from __future__ import print_function, unicode_literals
import getpass
import os
import tempfile


class PassphraseFile(object):
    """Context that provides a passphrase in a temporary file.

    The passphrase is queried from the user. Avoid repeated passphrase
    queries by caching entered passphrases.
    """

    passphrase_file = None
    passphrase_cache = {}

    def __init__(self, for_what, double_entry=False):
        """Setup passphrase file wrapper. `for_what` describes the object that
        is protected via this passphrase (presented to the user during
        passphrase entry). If `double_entry` is True, the passphrase must be
        entered twice.
        """
        self.for_what = for_what
        self.double_entry = double_entry

    def __enter__(self):
        self.passphrase_file = tempfile.NamedTemporaryFile(
            prefix='passphrase.', delete=False)
        print(self._get_passphrase(), file=self.passphrase_file)
        self.passphrase_file.close()
        return self.passphrase_file.name

    def _get_passphrase(self):
        if self.for_what in self.passphrase_cache:
            return self.passphrase_cache[self.for_what]
        if self.double_entry:
            phrase = self.ask_passphrase_twice()
        else:
            phrase = self.ask_passphrase_once()
        self.passphrase_cache[self.for_what] = phrase
        return phrase

    def ask_passphrase_once(self):
        return getpass.getpass('Enter passphrase for %s: ' % self.for_what)

    def ask_passphrase_twice(self):
        phrase1 = getpass.getpass('Enter passphrase for %s: ' % self.for_what)
        if len(phrase1) < 20:
            raise RuntimeError(
                'passphrase must be at least 20 characters long')
        phrase2 = getpass.getpass(
            'Enter passphrase for %s again: ' % self.for_what)
        if phrase1 != phrase2:
            raise RuntimeError('passphrases do not match')
        return phrase1

    def __exit__(self, *_exc_args):
        os.unlink(self.passphrase_file.name)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt
"""Host class and related code to interface with target hosts."""

from __future__ import print_function, unicode_literals
import contextlib
import fcntl
import os


def escape(path):
    """Return a safely escaped version of path.

    Implementation inspired by re.escape.
    """
    s = list(path)
    for i, c in enumerate(path):
        if c in ('\\', '"', '$'):
            s[i] = '\\' + c
    return path[:0].join(['"'] + s + ['"'])


class Host(object):
    """Deploy to an individual remote host."""

    # until we know for better, we leave service_home replacements alone
    service_home = u'${service_home}'

    def __init__(self, fqdn, environment):
        self.environment = environment
        self.fqdn = fqdn
        self.name = self.fqdn.split('.')[0]
        self.components = []

    @contextlib.contextmanager
    def locked(self):
        with open('.batou-lock', 'a+') as lockfile:
            try:
                fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except IOError:
                raise RuntimeError(
                    'cannot create lock "%s": more than one instance running '
                    'concurrently?' % lockfile, lockfile)
            # publishing the process id comes handy for debugging
            lockfile.seek(0)
            lockfile.truncate()
            print(os.getpid(), file=lockfile)
            lockfile.flush()
            yield
            lockfile.seek(0)
            lockfile.truncate()

    def deploy(self):
        with self.locked():
            for component in self.components:
                component.deploy()



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

"""Batou unit and functional tests"""

import os.path
import unittest


class TestCase(unittest.TestCase):
    """TestCase base class with custom extensions."""

    def assertFileExists(self, filename):
        self.assertTrue(os.path.exists(filename),
                        'file %s does not exists but should' % filename)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
from batou.component import Component, Buildout
from batou.tests import TestCase
import mock
import os
import os.path
import shutil
import sysconfig
import tempfile


class ComponentTests(TestCase):

    def setUp(self):
        self.host = mock.Mock()
        self.config = dict()
        self.component = Component('haproxy', self.host, (), self.config)
        self.component.hooks['haproxy'] = mock.Mock()
        self.host.components = [self.component]
        self.component.environment.hosts = dict(host=self.host)
        self.fixture = os.path.dirname(__file__) + '/fixture/component'

    def test_general_attributes_initialized(self):
        self.component.service.base = '<basedir>'
        self.assertEqual('haproxy', self.component.name)
        self.assertIs(self.host, self.component.host)
        self.assertIs(self.host.environment, self.component.environment)
        self.assertIs(self.host.environment.service, self.component.service)
        self.assertTrue('<basedir>/work/haproxy', self.component.compdir)

    # component wire-up

    def test_find_components_only_returns_if_tag_matches_no_host_given(self):
        self.assertRaises(RuntimeError, self.component.find_hooks, 'asdf')
        self.assertEqual(
            [self.component.hooks['haproxy']],
            list(self.component.find_hooks('haproxy')))

    def test_find_components_only_returns_if_tag_and_host_matches(self):
        self.assertRaises(RuntimeError,
                          self.component.find_hooks, 'haproxy', mock.Mock())
        self.assertEqual(
            [self.component.hooks['haproxy']],
            list(self.component.find_hooks('haproxy', self.host)))

    # convenience api

    def test_install_should_create_leading_dirs(self):
        with open('myfile', 'w') as f:
            print('hello world', file=f)
        self.component.install('myfile', 'tmp-a/b/c/d/myfile')
        self.assertFileExists('tmp-a/b/c/d/myfile')
        os.unlink('myfile')
        shutil.rmtree('tmp-a')

    def test_install_should_set_mode(self):
        with open('myfile1', 'w') as f:
            print('hello world', file=f)
        self.component.install('myfile1', 'myfile2', mode=0o654)
        self.assertEqual(os.stat('myfile2').st_mode & 0o7777, 0o654)
        os.unlink('myfile1')
        os.unlink('myfile2')

    def test_install_should_do_nothing_if_identical(self):
        with tempfile.NamedTemporaryFile(prefix='install') as tf:
            shutil.copyfile('%s/haproxy.cfg' % self.fixture, tf.name)
            os.utime(tf.name, (1, 1))
            self.component.install('%s/haproxy.cfg' % self.fixture, tf.name)
            self.assertEqual(1, os.stat(tf.name).st_mtime)
            self.assertEqual(set(), self.component.changed_files)

    def test_install_should_record_changed_file(self):
        with tempfile.NamedTemporaryFile(prefix='install') as tf:
            self.component.install('%s/haproxy.cfg' % self.fixture, tf.name)
            self.assertEqual(set([tf.name]), self.component.changed_files)

    def test_chdir(self):
        olddir = os.getcwd()
        with self.component.chdir('/tmp'):
            self.assertEqual('/tmp', os.getcwd())
        self.assertEqual(olddir, os.getcwd())

    # template api

    def test_expand(self):
        self.assertEqual('hello haproxy',
                         self.component.expand('hello ${component.name}'))

    def test_template_should_do_nothing_if_identical(self):
        with tempfile.NamedTemporaryFile(prefix='install') as tf:
            with open(tf.name, 'w') as interpolated:
                interpolated.write("""\
frontend
        name haproxy
""")
            os.utime(tf.name, (1, 1))
            self.component.template('%s/haproxy.cfg' % self.fixture, tf.name)
            self.assertEqual(1, os.stat(tf.name).st_mtime,
                             'file has been touched')
            self.assertEqual(set(), self.component.changed_files)

    def test_template_should_record_changed_file(self):
        with tempfile.NamedTemporaryFile(prefix='install') as tf:
            self.component.template('%s/haproxy.cfg' % self.fixture, tf.name)
            self.assertEqual(set([tf.name]), self.component.changed_files)


class BuildoutTests(TestCase):

    @mock.patch('os.symlink')
    def test_determine_python_config(self, symlink):
        c = Buildout(mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock())
        c.python = 'python%s' % sysconfig.get_python_version()
        c._install_python_config()
        self.assertTrue(symlink.called)



# Copyright (c) gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
import socket
import os
import subprocess
import urlparse


def string_list(config_value):
    """Explode comma-separated config value."""
    config_value = config_value.strip()
    if not config_value:
        return []
    return [x.strip() for x in config_value.split(',')]


def convert_type(value, datatype):
    """Allow str, int, float, and bool values in config files."""
    if datatype == 'str':
        return str(value)
    if datatype == 'int':
        return int(value)
    if datatype == 'float':
        return float(value)
    if datatype == 'bool':
        if value.lower() in ('false', 'no', 'n', 'off'):
            return False
        return bool(value)
    if datatype == 'list':
        return string_list(value)
    raise ValueError('unknown data type', datatype)


def check_shared_object(sopath):
    """Return True is shared object is link-consistent."""
    if not os.path.exists(sopath):
        return False
    deps = subprocess.check_output(['ldd', sopath])
    return '=> not found' not in deps


def prepend_env(envvar, string):
    """Put a `string` in front of a PATH-like `envvar`.

    The elements of the `envvar` are assumed to be colon-separated. If the
    `envvar` does not exist before, it is created just with `string`.
    """
    try:
        if string not in os.environ[envvar]:
            os.environ[envvar] = '%s:%s' % (string, os.environ[envvar])
    except KeyError:
        os.environ[envvar] = string


def host_from_uri(uri):
    """Extract host part from URI."""
    host = urlparse.urlsplit(uri).netloc
    host = host.split(':')[0]
    if '@' in host:
        host = host.split('@')[1]
    return host


def resolve(address):
    if ':' in address:
        host, port = address.split(':')
    else:
        host, port = address, None
    try:
        address = socket.gethostbyname(host)
    except socket.gaierror as e:
        raise socket.gaierror('%s (%s)' % (str(e), host))
    if port:
        address += ':%s' % port
    return address


def max_mtime(files):
    """Returns the latest mtime for a list of files"""
    return max(os.stat(f).st_mtime for f in files)


class Address(object):
    """An internet service address that can be listened and connected to.

    The expected constructor address is expected to be the address that can be
    connected to. The listen address will be computed automatically.

    """

    def __init__(self, connect_address, port=None):
        if ':' in connect_address:
            connect, port = connect_address.split(':')
        else:
            connect = connect_address
        listen = resolve(connect)
        self.listen = NetLoc(listen, str(port))
        self.connect = NetLoc(connect, str(port))


class NetLoc(object):

    def __init__(self, host, port=None):
        self.host = host
        self.port = port

    def __str__(self):
        result = self.host
        if self.port:
            result += ':' + self.port
        return result


class Hook(object):
    pass



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from batou.component import Component
from batou.service import ServiceConfig
import os
import os.path
import pwd
import unittest


class BrokenConfigTests(unittest.TestCase):

    def test_parse_nonexisting_file_raises_ioerror(self):
        with self.assertRaises(IOError):
            ServiceConfig(os.path.dirname(__file__) +
                          '/no/such/file.cfg', ['foo'])


class ConfigTestsBasicScenario(unittest.TestCase):

    def setUp(self):
        self.config = ServiceConfig(os.path.dirname(__file__) +
                   '/fixture/basic_service', ['dev'])
        self.config.scan()
        for env in self.config.service.environments.values():
            self.config.configure_components(env)
        self.service = self.config.service

    def test_service_options_are_set(self):
        self.assertTrue(self.service.base.endswith(
            '/fixture/basic_service'))

    def test_components_are_fully_loaded(self):
        self.assertEquals(['zeo', 'zope'], sorted(self.service.components))

    def test_dev_is_loaded_but_production_is_not(self):
        self.assertEquals(['dev'],
                          sorted(self.service.environments))

    def test_production_environment_is_loaded(self):
        self.config.environments = set(['production'])
        self.config.scan()
        self.config.configure_components(
            self.config.service.environments['production'])
        production = self.service.environments['production']
        self.assertEquals('alice', production.service_user)
        self.assertEquals('production', production.branch)
        self.assertEquals('example.com', production.host_domain)

        self.assertEquals(['host1.example.com', 'host2.example.com',
                           'host3.example.com'],
                          sorted(production.hosts))

        host1 = production.hosts['host1.example.com']
        self.assertEqual('host1', host1.name)
        self.assertEqual('host1.example.com', host1.fqdn)
        self.assertEqual(1, len(host1.components))
        zope = host1.components[0]
        self.assertIsInstance(zope, Component)
        self.assertEqual('zope', zope.name)

    def test_dev_environment_is_loaded(self):
        dev = self.service.environments['dev']
        self.assertEquals(pwd.getpwuid(os.getuid()).pw_name, dev.service_user)
        self.assertEquals('default', dev.branch)
        self.assertEquals(None, dev.host_domain)

        self.assertEquals(['localhost'], sorted(dev.hosts))

        localhost = dev.hosts['localhost']
        self.assertEqual('localhost', localhost.name)
        self.assertEqual('localhost', localhost.fqdn)
        components = list(localhost.components)
        components.sort(key=lambda x:x.name)
        self.assertEqual(['zeo', 'zope'], [x.name for x in components])
        zeo = components[0]
        self.assertIsInstance(zeo, Component)
        self.assertEqual('zeo', zeo.name)

    def test_component_has_features_set(self):
        dev = self.service.environments['dev']
        self.assertEquals(['test'],
                          dev.hosts['localhost'].components[1].features)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from .component import Component
from .host import Host
from batou import Environment
import batou.utils
import collections
import ConfigParser
import glob
import os
import os.path


class ServiceConfig(object):
    """A service configuration.

    Handles the on-disk format of service, component, and environment
    parameters.

    Two phases: scan, configure

    Scan -- discovers all elements (environments, hosts, components) but
    doesn't create an actual configuration

    Configure -- turns all discovered elements into an actual
    configuration that can be deployed.

    This two-phase approach is required to allow better instrumentation
    of the loading process while integrating with Fabric.
    """

    def __init__(self, basedir, environments=[]):
        self.environments = set(environments)
        self.basedir = os.path.abspath(basedir)
        if not os.path.exists(self.basedir):
            raise IOError('No such file or directory: %r' % self.basedir)
        self.component_pattern = (self.basedir + '/components/*/component.py')
        self.environment_pattern = (self.basedir + '/environments/*.cfg')

        service = Service()
        service.base = self.basedir
        self.service = service

    def scan(self):
        self.scan_components()
        self.scan_environments()

    def scan_components(self):
        for filename in glob.glob(self.component_pattern):
            for factory, name in Component.from_file(filename):
                self.service.components[name] = factory

    def scan_environments(self):
        """Set up basic environment objects for all configs found."""
        existing_environments = glob.glob(self.environment_pattern)
        for pattern in [self.service.base + '/environments/', '.cfg']:
            existing_environments = map(lambda x: x.replace(pattern, ''),
                                           existing_environments)
        self.existing_environments = set(existing_environments)

        if self.environments:
            self.environments = self.existing_environments & self.environments
        else:
            self.environments = self.existing_environments
        for environment in self.environments:
            self.load_environment(environment)

    def load_environment(self, environment):
        """Instantiate environment object structure from config."""
        config = ConfigParser.SafeConfigParser()
        config.read('%s/environments/%s.cfg' % (
            self.service.base, environment))
        try:
            env = Environment(environment, self.service)
            env.configure(dict(config.items('environment')))
        except Exception, e:
            print "Problem loading environment %s: %r (ignored)" % (
                environment, e)
            return
        self.load_hosts(env, config)
        self.service.environments[env.name] = env

    def load_hosts(self, env, config):
        for hostname in config.options('hosts'):
            fqdn = env.normalize_host_name(hostname)
            env.hosts[fqdn] = host = Host(fqdn, env)
            # load components for host
            components = collections.OrderedDict()
            for component in batou.utils.string_list(
                    config.get('hosts', hostname)):
                feature = None
                if ':' in component:
                    component, feature = component.split(':')
                components.setdefault(component, [])
                if feature:
                    components[component].append(feature)
            self.assemble_host_components(host, config, components)

    def assemble_host_components(self, host, config, components):
        """Create all components for a host and do the wire-up."""
        for component, features in components.items():
            factory = self.service.components[component]
            component_config = {}
            if config.has_section('component:%s' % component):
                component_config.update(
                    dict(config.items('component:%s' % component)))
            if not features:
                features = factory.features
            host.components.append(factory(
                component, host, features, component_config))

    def configure_components(self, environment):
        for host in environment.hosts.values():
            for component in host.components:
                component._configure()


class Service(object):

    base = None

    def __init__(self):
        self.components = {}
        self.environments = {}



# Copyright (c) gocept gmbh & co. kg
# See also LICENSE.txt

from batou.template import TemplateEngine
import batou.utils
import contextlib
import filecmp
import glob
import os
import os.path
import re
import shutil
import subprocess


class Component(object):

    configured = False
    defdir = None  # The component's definition directory
    template_format = 'mako'
    features = ()

    @classmethod
    def from_file(cls, filename):
        g = l = {}
        g.update(globals())
        oldcwd = os.getcwd()
        defdir = os.path.dirname(filename)
        os.chdir(defdir)
        execfile(filename, g, l)
        for candidate in l.values():
            if candidate in globals().values():
                continue
            if not isinstance(candidate, type):
                continue
            if issubclass(candidate, Component):
                candidate.defdir = defdir
                yield candidate, candidate.__name__.lower()
        os.chdir(oldcwd)

    def __init__(self, name, host, features, config, template_format=None):
        self.host = host
        self.environment = host.environment
        self.service = host.environment.service
        self.name = name
        self.config = config
        self.hooks = {}
        self.setup_hooks()
        self.features = features
        self.changed_files = set()
        if template_format:
            self.template_format = template_format
        self.template_engine = TemplateEngine.get(self.template_format)

    def setup_hooks(self):
        """Declare this component's hooks."""

    def configure(self):
        """Prepare this component's configuration.

        Perform any processing of configuration options, information
        retrieval from other components, template expansion, etc, so
        that we get ready to deploy this component.
        """
        pass  # override in subclasses

    def _configure(self):
        if self.configured:
            return
        self.configured = True
        self.configure()

    # Convenience API

    @property
    def compdir(self):
        return '%s/work/%s' % (self.service.base, self.name)

    def find_hooks(self, required_name, host=None):
        hooks = []
        required_name = re.compile(required_name)
        for candidate_host in self.host.environment.hosts.values():
            if host and candidate_host is not host:
                continue
            for component in candidate_host.components:
                for name, hook in component.hooks.items():
                    if required_name.match(name):
                        # Ensure that if some component is requested, we will
                        # always return components that have been configured
                        # previously.
                        component._configure()
                        if name not in component.hooks.keys():
                            # Configuration may have caused hook to be
                            # disabled.
                            continue
                        hook._name = name
                        hooks.append(hook)
        if not hooks:
            raise RuntimeError('No hook found: name=%s, host=%s' %
                               (required_name.pattern, host))
        return iter(hooks)

    def config_attr(self, name, datatype='str'):
        value = self.config.get(name, getattr(self, name, ''))
        value = self.expand(str(value))
        value = batou.utils.convert_type(value, datatype)
        setattr(self, name, value)
        return value

    def cmd(self, cmd):
        self.log('%s: cmd: %s' % (os.getcwd(), cmd))
        output = subprocess.check_output([cmd], shell=True)
        self.log(output)
        return output

    def changed_file(self, filename):
        """Add `filename` to the list of changed files."""
        self.changed_files.add(filename)

    def create_leading_dirs(self, targetfile):
        try:
            os.makedirs(os.path.dirname(targetfile))
        except OSError:
            # ignore failures as long as we get the file in place
            pass

    def install(self, source, target, mode=None, makedirs=True):
        """Copy file, chmod optionally, create leading directories."""
        if makedirs:
            self.create_leading_dirs(target)
        if not os.path.exists(target) or not filecmp.cmp(source, target):
            shutil.copyfile(source, target)
            self.changed_file(target)
        if mode and os.stat(target).st_mode & 0o7777 != mode:
            os.chmod(target, mode)
            self.changed_file(target)

    @contextlib.contextmanager
    def chdir(self, directory):
        """Execute associated with block in `directory`."""
        olddir = os.getcwd()
        os.chdir(directory)
        yield
        os.chdir(olddir)

    def log(self, msg):
        print msg.strip()

    # Template API

    @property
    def _template_args(self):
        args = dict(
            host=self.host,
            environment=self.environment,
            service=self.environment.service,
            component=self)
        return args

    def template(self, filename, target):
        """Render template from `filename` into `target` file."""
        changed = self.template_engine.template(filename, target,
                                                self._template_args)
        if changed:
            self.changed_file(target)

    def expand(self, templatestr):
        """Return rendered string template as string."""
        return self.template_engine.expand(templatestr, self._template_args)

    # Multi-step deployment strategy

    def _update_compdir(self):
        if not os.path.exists(self.compdir):
            os.makedirs(self.compdir)
        for dirpath, dirnames, filenames in os.walk(self.defdir):
            working_dirpath = dirpath.replace(self.defdir, self.compdir)
            for dirname in dirnames:
                dirname = os.path.join(working_dirpath, dirname)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                    self.changed_file(dirname)
            for filename in filenames:
                source_filename = os.path.join(dirpath, filename)
                dest_filename = os.path.join(working_dirpath, filename)
                # Check whether file is up to date at target already.
                if not os.path.exists(dest_filename):
                    pass
                elif (open(source_filename, 'r').read() !=
                      open(dest_filename, 'r').read()):
                    pass
                elif (os.stat(dest_filename).st_mode !=
                      os.stat(source_filename).st_mode):
                    pass
                else:
                    # The target file is up-to-date.
                    continue
                shutil.copy(source_filename, dest_filename)
                self.changed_file(dest_filename)

    def deploy(self):
        """Run all methods annotated with @step(n) in sequence."""
        self.log('deploy: %s' % self.__class__.__name__)
        self._update_compdir()
        os.chdir(self.compdir)
        for step in self.steps():
            self.log('step: %s' % step.__name__)
            step()

    def steps(self):
        """Return ordered list of deployment steps."""
        steps = []
        for name in dir(self):
            f = getattr(self, name)
            if not callable(f):
                continue
            if not getattr(f, 'step', False):
                continue
            steps.append(f)
        steps.sort(key=lambda x: x.step)
        return steps


def step(x):
    def decorate_step(f):
        f.step = x
        return f
    return decorate_step


class Buildout(Component):

    python = 'python2.6'    # which python to use for virtualenv
    executable = None       # path to executable generated by virtualenv
    profile = 'base'

    def configure(self):
        self.config_attr('profile')
        self.find_links = ''

    @property
    def executable(self):
        return 'bin/%s' % self.python

    @property
    def py_version(self):
        return self.python.strip('python')

    def _install_virtualenv(self):
        self.cmd('virtualenv --no-site-packages --python %s .' % self.python)

    def _python_config_candidates(self):
        """List possible file names for python-config."""
        return [cand % dict(py_version=self.py_version) for cand in (
            'python%(py_version)s-config',
            'python-config-%(py_version)s',
            'python-config%(py_version)s',
        )]

    def _install_python_config(self):
        """Create symlink to correct python-config in virtualenv."""
        script_path = subprocess.check_output([
            "%s -c 'from distutils import sysconfig; "
            "print(sysconfig.EXEC_PREFIX)'" % self.python], shell=True).strip()
        for candidate in self._python_config_candidates():
            candidate = os.path.join(script_path, 'bin', candidate)
            if not os.path.exists(candidate):
                continue
            os.symlink(candidate, 'bin/python-config')
            break
        else:
            raise RuntimeError('cannot find suitable python-config',
                               self._python_config_candidates())

    @step(1)
    def install(self):
        """Basic prerequisites: python virtual and supporting files."""
        if not os.path.exists(self.executable):
            self._install_virtualenv()
        if not os.path.islink('bin/python-config'):
            try:
                os.unlink('bin/python-config')
            except OSError:
                pass
            self._install_python_config()

    @step(2)
    def generate_config(self):
        try:
            secrets = self.find_hooks('secrets').next()
            self.find_links = secrets.get('buildout', 'find-links')
        except Exception:
            print('missing [buildout]find-links in secrets file, ignoring')
        self.template('buildout.cfg.in', target='buildout.cfg')

    def _bootstrap_required(self):
        """Returns True if we must run buildout."""
        if not os.path.exists('bin/buildout'):
            return True
        buildout_mtime = os.stat('bin/buildout').st_mtime
        if (buildout_mtime > (os.stat(self.executable).st_mtime) and
            buildout_mtime > batou.utils.max_mtime(
                glob.glob('buildout.cfg') + glob.glob('profiles/*'))):
            return False
        return True

    @step(3)
    def bootstrap(self):
        if self._bootstrap_required():
            self.cmd('%s bootstrap.py' % self.executable)

    @step(4)
    def buildout(self):
        output = self.cmd('bin/buildout -t 15')
        # Special case for handling mr.developer exits
        # can be removed once https://bugs.launchpad.net/zc.buildout/+bug/962169
        # is fixed.
        if 'mr.developer: There have been errors, see messages' in output:
            raise RuntimeError('mr.developer encountered an error.')



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
from batou.utils import host_from_uri, resolve, string_list, convert_type
import mock
import unittest


class HostFromUriTests(unittest.TestCase):

    def test_source_host_from_simple_url(self):
        self.assertEqual(
            'example.com', host_from_uri('http://example.com'))

    def test_source_host_ignores_port(self):
        self.assertEqual(
            'example.com', host_from_uri('http://example.com:80'))

    def test_source_host_ignores_user(self):
        self.assertEqual(
            'example.com', host_from_uri('http://user@example.com'))

    def test_source_host_ignores_user_and_port(self):
        self.assertEqual(
            'example.com', host_from_uri('http://user@example.com:80'))


class ResolveTests(unittest.TestCase):

    @mock.patch('socket.gethostbyname')
    def test_host_without_port_resolves(self, ghbn):
        ghbn.return_value = '127.0.0.1'
        self.assertEqual('127.0.0.1', resolve('localhost'))

    @mock.patch('socket.gethostbyname')
    def test_host_with_port_resolves_and_keeps_port(self, ghbn):
        ghbn.return_value = '127.0.0.1'
        self.assertEqual('127.0.0.1:8080', resolve('localhost:8080'))


class StringList(unittest.TestCase):

    def test_empty_string_returns_empty_list(self):
        self.assertEquals([], string_list(''))
        self.assertEquals([], string_list('  '))

    def test_comma_separates_items(self):
        self.assertEquals(['a', 'b', 'c'], string_list('a,b,c'))

    def test_whitespace_is_stripped(self):
        self.assertEquals(['a', 'b', 'c'], string_list(' a , b , c '))

    def test_single_item_returns_list_with_single_item(self):
        self.assertEquals(['a'], string_list('a'))

    def test_single_item_is_stripped(self):
        self.assertEquals(['a'], string_list('a '))


class ConvertTypeTest(unittest.TestCase):

    def test_str(self):
        self.assertEqual('1', convert_type(1, 'str'))

    def test_int(self):
        self.assertEqual(2, convert_type('2', 'int'))

    def test_float(self):
        self.assertEqual(3.6, convert_type('3.6', 'float'))

    def test_bool(self):
        self.assertEqual(True, convert_type('yes', 'bool'))
        self.assertEqual(True, convert_type('on', 'bool'))
        self.assertEqual(False, convert_type('n', 'bool'))
        self.assertEqual(False, convert_type('no', 'bool'))
        self.assertEqual(False, convert_type('false', 'bool'))
        self.assertEqual(False, convert_type('off', 'bool'))

    def test_list(self):
        self.assertEqual(['one', 'two'], convert_type('one, two', 'list'))



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
from batou.host import Host
from batou.tests import TestCase
import multiprocessing


class InterlockedComponent(object):

    def __init__(self, lock1, lock2):
        self.lock1 = lock1
        self.lock2 = lock2

    def deploy(self):
        self.lock1.release()
        self.lock2.acquire()
        self.lock2.release()


class HostTest(TestCase):

    def test_host_should_create_lock(self):
        step1 = multiprocessing.Lock()
        step1.acquire()
        step2 = multiprocessing.Lock()
        step2.acquire()
        firsthost = Host('h1', 'env')
        firsthost.components.append(InterlockedComponent(step1, step2))
        t = multiprocessing.Process(target=firsthost.deploy)
        t.start()
        try:
            secondhost = Host('h2', 'env')
            step1.acquire()
            with self.assertRaises(RuntimeError):
                secondhost.deploy()
        finally:
            step1.release()
            step2.release()
            t.join()



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
from batou.secrets import SecretsFile
from batou.tests import TestCase
import os
import os.path
import shutil
import tempfile


class SecretsFileTests(TestCase):

    def setUp(self):
        self.cleartext = (os.path.dirname(__file__) +
                          '/fixture/secrets/cleartext.cfg')
        self.encrypted = (os.path.dirname(__file__) +
                          '/fixture/secrets/encrypted.cfg')
        pf = tempfile.NamedTemporaryFile(prefix='passphrase.', delete=False)
        print('SecretTestPassphrase', file=pf)
        pf.close()
        self.passphrase = pf.name

    def tearDown(self):
        os.unlink(self.passphrase)

    def test_decrypt(self):
        with SecretsFile(self.encrypted, self.passphrase) as secrets:
            with open(self.cleartext) as cleartext:
                self.assertEqual(cleartext.read(), secrets.read())

    def test_decrypt_wrong_passphrase(self):
        with open(self.passphrase, 'w') as pf:
            print('wrong passphrase wrong passphrase', file=pf)
            pf.flush()
        with self.assertRaises(RuntimeError):
            with SecretsFile(self.encrypted, self.passphrase) as secrets:
                print(secrets.read())

    def test_write_should_fail_unless_write_locked(self):
        with SecretsFile(self.encrypted, self.passphrase) as secrets:
            self.assertRaises(RuntimeError, secrets.write, 'dummy')

    def test_open_nonexistent_file_for_read_should_fail(self):
        with self.assertRaises(IOError):
            with SecretsFile('/no/such/file', self.passphrase):
                pass

    def test_open_nonexistent_file_for_write_should_create_empty_file(self):
        tf = tempfile.NamedTemporaryFile(prefix='new_encrypted.')
        tf.close()  # deletes file
        with SecretsFile(tf.name, self.passphrase, write_lock=True) as sf:
            self.assertEqual('', sf.read())
            self.assertFileExists(sf.encrypted_file)
        os.unlink(tf.name)

    def test_write(self):
        with tempfile.NamedTemporaryFile(prefix='new_encrypted.') as tf:
            shutil.copy(self.encrypted, tf.name)
            with SecretsFile(tf.name, self.passphrase, write_lock=True) as sf:
                sf.write('some new file contents\n')
            with open(self.encrypted, 'rb') as old:
                with open(tf.name, 'rb') as new:
                    self.assertNotEqual(old.read(), new.read())
            self.assertNotEqual(0, os.stat(tf.name).st_size)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
from batou.template import TemplateEngine
from batou.tests import TestCase
import collections
import jinja2
import mock
import os
import os.path
import shutil
import tempfile


Server = collections.namedtuple('Server', ['name', 'address'])


class ComponentTemplateTests(TestCase):

    def setUp(self):
        self.host = mock.Mock()
        self.config = dict()
        self.servers = [Server('s1', '1.2.3.4'), Server('s2', '2.3.4.5')]
        self.hello = 'world'
        self.fixture = '%s/fixture/template' % os.path.dirname(__file__)

    def test_unknown_template_engine(self):
        with self.assertRaises(NotImplementedError):
            TemplateEngine.get('foo')

    def test_mako_template_str(self):
        tmpl = TemplateEngine.get('mako')
        self.assertEqual('hello world',
                         tmpl.expand('hello ${hello}', self.__dict__))

    def test_jinja2_template_str(self):
        tmpl = TemplateEngine.get('jinja2')
        self.assertEqual('hello world',
                         tmpl.expand('hello {{hello}}', self.__dict__))

    def test_mako_dollar_pseudoescape(self):
        tmpl = TemplateEngine.get('mako')
        self.assertEqual('${PATH}', tmpl.expand('${d}{PATH}', {}))

    def _template_runner(self, template_format, source):
        tmpl = TemplateEngine.get(template_format)
        with tempfile.NamedTemporaryFile(prefix='out') as target:
            changed = tmpl.template(source, target.name, self.__dict__)
            self.assertTrue(changed)
            with open('%s/haproxy.cfg' % self.fixture) as ref:
                with open(target.name) as tf:
                    self.assertMultiLineEqual(ref.read(), tf.read())

    def test_mako_template_file(self):
        self._template_runner('mako', '%s/haproxy.cfg.mako' % self.fixture)

    def test_jinja2_template_file(self):
        self._template_runner('jinja2', '%s/haproxy.cfg.jinja2' % self.fixture)

    def test_expand_should_return_false_if_unchanged(self):
        tmpl = TemplateEngine.get('jinja2')
        source = '%s/haproxy.cfg' % self.fixture
        with tempfile.NamedTemporaryFile(prefix='out') as target:
            shutil.copyfile(source, target.name)
            #import pdb; pdb.set_trace()
            changed = tmpl.template(source, target.name, self.__dict__)
            self.assertFalse(changed)


    def test_jinja2_unknown_variable_should_fail(self):
        tmpl = TemplateEngine.get('jinja2')
        with self.assertRaises(jinja2.UndefinedError):
            tmpl.expand('unknown variable {{foo}}', {})



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

import unittest
import mock
from batou.local import main


class LocalTests(unittest.TestCase):

    @mock.patch('sys.argv')
    @mock.patch('sys.stderr')
    def test_no_args_exits_with_error(self, stderr, argv):
        with self.assertRaises(SystemExit):
            main()



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
from batou.utils import Address, Hook
from batou.component import Component, step


class HAProxy(Component):
    """Load balancing component.

    Configuration attributes::
        [component:haproxy]
        haproxy_cfg = PATH
            path name of the configuration file
        restart = BOOL
            whether to call /etc/init.d/haproxy

    Expects hooks (given by backend_hooks) with the attributes:

    - server_name (string)
    - address (Address object)
    - backend (string)

    Provides hook 'haproxy:frontend' which has the following attributes:

    - address (Address object)

    """

    haproxy_cfg = '/etc/haproxy.cfg'
    restart = 'true'
    address = '${host.fqdn}:8002'
    backend_hooks = ()

    def setup_hooks(self):
        self.hooks['haproxy:frontend'] = Hook()

    def configure(self):
        self.config_attr('haproxy_cfg')
        self.config_attr('restart', 'bool')
        self.address = Address(self.config_attr('address'))
        self.hooks['haproxy:frontend'].address = self.address

        self.backends = {}
        for hook_name in self.backend_hooks:
            for hook in self.find_hooks(hook_name):
                servers = self.backends.setdefault(hook.backend, [])
                servers.append(hook)

    @step(1)
    def update_config(self):
        self.template('haproxy.cfg', target=self.haproxy_cfg)

    @step(2)
    def restart_haproxy(self):
        if not self.restart:
            self.log('skipping haproxy restart')
            return
        # restart instead of reload due to kernel bug
        self.cmd('sudo /etc/init.d/haproxy restart')



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

# This should be only one line. If it must be multi-line, indent the second
# line onwards to keep the PKG-INFO file format intact.
"""A utility for automating multi-host, multi-environment software builds and deployments.
"""

from setuptools import setup, find_packages
import glob
import os.path


def project_path(*names):
    return os.path.join(os.path.dirname(__file__), *names)


version = '0.1.4'

setup(
    name='batou',
    version=version,
    install_requires=[
        'distribute',
        'Fabric',
        'Mako',
        'Jinja2',
        'mock',
    ],
    extras_require={
        'test': [
        ],
    },
    entry_points="""
        [console_scripts]
            batou-remote = batou.remote:main
            batou-local = batou.local:main
            secretsedit = batou.secrets:edit
    """,
    author='gocept <mail@gocept.com>',
    author_email='mail@gocept.com',
    license='ZPL',
    url='https://projects.gocept.com/projects/batou/',
    keywords='deployment',
    classifiers="""\
License :: OSI Approved :: Zope Public License
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.6
Programming Language :: Python :: 2.7
Programming Language :: Python :: 2 :: Only
"""[:-1].split('\n'),
    description=__doc__.strip(),
    long_description='\n\n'.join(open(project_path(name)).read() for name in (
            'README.txt',
            'HACKING.txt',
            'CHANGES.txt',
            )),

    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    data_files=[('', glob.glob(project_path('*.txt')))],
    zip_safe=False,
    test_suite='batou.tests',
)



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from __future__ import print_function, unicode_literals
from batou.service import Service, ServiceConfig
import batou.tests
import os.path
import unittest


class ServiceTests(unittest.TestCase):

    def test_datastructures_init(self):
        service1 = Service()
        service2 = Service()
        self.assertIsNot(service1.components, service2.components)
        self.assertIsNot(service1.environments, service2.environments)

class ServiceConfigTest(batou.tests.TestCase):

    def setUp(self):
        self.fixture = os.path.dirname(__file__) + '/fixture/ordering'
        self.config = ServiceConfig(self.fixture, ['test'])
        self.config.scan()

    def test_config_should_define_component_order(self):
        env = self.config.service.environments['test']
        self.config.configure_components(env)
        components = env.hosts['localhost'].components
        self.assertEqual([c.name for c in components], ['c1', 'c2', 'c3'])



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from .environment import Environment
from .service import Service, ServiceConfig
from .remote import FabricTask
import inspect
import os.path


def load():
    # Helper to make fabfile more digestable
    caller = inspect.getouterframes(inspect.currentframe())[1]
    base_dir = os.path.dirname(caller[0].f_globals['__file__'])
    config = ServiceConfig(base_dir)
    config.scan()
    fab_commands = caller[0].f_globals
    fab_commands.update(FabricTask.from_config(config))



# Copyright (c) 2012 gocept gmbh & co. kg
# See also LICENSE.txt

from .service import ServiceConfig
import argparse
import locale


def main():
    locale.setlocale(locale.LC_ALL, 'en_US.utf8')
    parser = argparse.ArgumentParser(
        description=u'Execute batou components locally.')
    parser.add_argument(
        'environment', help='Environment to deploy.',
        type=lambda x:x.replace('.cfg', ''))
    parser.add_argument(
        'hostname', help='Host to deploy.')
    args = parser.parse_args()

    config = ServiceConfig('.', [args.environment])
    config.scan()
    environment = config.service.environments[args.environment]
    config.configure_components(environment)
    environment.get_host(args.hostname).deploy()
